import importlib
from abc import ABC, abstractmethod
from enum import IntEnum
from functools import reduce

import pandas as pd


class TermEnum(IntEnum):
    """
    Term types.

    - CAT: categorical without levels specified
    - LVL: categorical with levels specified
    - CON: continuous
    - FUN: continuous with function transformations
    - INT: intercept
    """
    CAT = 1
    LVL = 2
    CON = 3
    FUN = 4
    INT = 5

    @staticmethod
    def get_extractor(record, term):
        """
        Gets the associated extractor based on the specified term.

        :param record: Dictionary.
        :param term: Model term.
        :return: Extractor.
        """

        if term.startswith('C'):
            return LvlExtractor(record, term)
        elif '[' in term and ']' in term:
            return CatExtractor(record, term)
        elif 'Intercept' == term:
            return IntExtractor(record, term)
        elif '(' in term and ')' in term:
            return FunExtractor(record, term)
        else:
            return ConExtractor(record, term)


class Extractor(ABC):
    """
    Extractor to get value based on model term.
    """

    def __init__(self, record, term, term_type):
        """
        ctor.

        :param: Dictionary.
        :term: Model term.
        :term_type: Type of term.
        :return: None
        """
        self._record = record
        self._term = term
        self._type = term_type

    def __repr__(self):
        return f'{self.__class__.__name__}[term={self._term}, type={self._type.name}]'

    @property
    @abstractmethod
    def value(self):
        """
        Gets the extracted value.
        """
        pass


class CatExtractor(Extractor):
    """
    Categorical extractor (no levels).
    """

    def __init__(self, record, term):
        """
        ctor.

        :param record: Dictionary.
        :param term: Model term.
        :return: None.
        """
        super().__init__(record, term, TermEnum.CAT)

    @property
    def value(self):
        idx = self._term.index('[')
        x_name = self._term[0:idx]

        if x_name not in self._record or self._record[x_name] is None:
            return None

        lhs = self._term.rindex('[') + 1
        rhs = self._term.rindex(']')
        x_val = self._term[lhs:rhs]
        x_val = x_val.replace('T.', '')

        if self._record[x_name] == x_val:
            return 1.0
        return 0.0


class LvlExtractor(Extractor):
    """
    Categorical extractor (with levels).
    """

    def __init__(self, record, term):
        """
        ctor.

        :param record: Dictionary.
        :param term: Model term.
        :return: None.
        """
        super().__init__(record, term, TermEnum.LVL)

    @property
    def value(self):
        lhs = self._term.index('(') + 1
        rhs = self._term.index(',')
        x_name = self._term[lhs:rhs]

        if x_name not in self._record or self._record[x_name] is None:
            return None

        lhs = self._term.rindex('[') + 1
        rhs = self._term.rindex(']')
        x_val = self._term[lhs:rhs]
        x_val = x_val.replace('T.', '')

        if self._record[x_name] == x_val:
            return 1.0
        return 0.0


class ConExtractor(Extractor):
    """
    Continuous extractor (no functions).
    """

    def __init__(self, record, term):
        """
        ctor.

        :param record: Dictionary.
        :param term: Model term.
        :return: None.
        """
        super().__init__(record, term, TermEnum.CON)

    @property
    def value(self):
        return self._record[self._term] if self._term in self._record else None


class IntExtractor(Extractor):
    """
    Intercept extractor. Always returns 1.0.
    """

    def __init__(self, record, term):
        """
        ctor.

        :param record: Dictionary.
        :param term: Model term.
        :return: None.
        """
        super().__init__(record, term, TermEnum.INT)

    @property
    def value(self):
        return 1.0


class FunExtractor(Extractor):
    """
    Continuous extractor (with functions defined).
    """

    def __init__(self, record, term):
        """
        ctor.

        :param record: Dictionary.
        :param term: Model term.
        :return: None.
        """
        super().__init__(record, term, TermEnum.FUN)

    # flake8: noqa: F841
    @property
    def value(self):
        lhs = self._term.rindex('(') + 1
        rhs = self._term.index(')')
        x_name = self._term[lhs:rhs]
        expression = f'{self._term[0:lhs]}val{self._term[rhs:]}'
        val = self._record[x_name] if x_name in self._record else None
        if pd.isna(val):
            return None

        if 'np.' in expression:
            np = importlib.import_module('numpy')
        v = eval(expression)

        if isinstance(v, np.generic):
            v = np.asscalar(v)
        return v


class InteractionExtractor(object):
    """
    Interaction extractor for interaction effects.
    """

    def __init__(self, record, terms):
        """
        ctor.

        :param record: Dictionary.
        :param terms: Model term (possibly with interaction effects).
        :return: None.
        """
        self._terms = terms
        extractors = [TermEnum.get_extractor(record, term) for term in terms.split(':')]
        values = [e.value for e in extractors]
        values = [v for v in values if pd.notna(v)]

        if len(values) != len(extractors):
            self.__value = None
        else:
            self.__value = reduce(lambda a, b: a * b, values)

    def __repr__(self):
        return f'{self.__class__.__name__}[terms={self._terms}]'

    @property
    def value(self):
        return self.__value



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_desc = fh.read()

setup(
    name='ydot',
    version='0.0.4',
    author='Jee Vang',
    author_email='vangjee@gmail.com',
    packages=find_packages(exclude=('*.tests', '*.tests.*', 'tests.*', 'tests')),
    description='R-like formulas for Spark Dataframes',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    url='https://github.com/oneoffcoder/pyspark-formula',
    keywords=' '.join(
        ['statistics', 'pyspark', 'formula', 'patsy', 'spark',
         'dataframe', 'regression', 'classification', 'data',
         'machine learning', 'artificial intelligence']),
    install_requires=['scipy', 'numpy', 'pandas', 'scikit-learn', 'pyspark', 'patsy'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Development Status :: 5 - Production/Stable'
    ],
    include_package_data=True,
    test_suite='nose.collector'
)



import importlib
from itertools import product

import pandas as pd
from patsy.highlevel import dmatrices
from pyspark import Row

from ydot.formula import InteractionExtractor


def get_profile(sdf):
    """
    Gets the field profiles of the specified Spark dataframe.

    :param sdf: Spark dataframe.
    :return: Dictionary.
    """
    dtypes = {k: v for k, v in sdf.dtypes}
    cat_types = sdf.rdd \
        .map(lambda r: r.asDict()) \
        .flatMap(lambda r: [((k, r[k]), 1) for k, v in dtypes.items() if v == 'string']) \
        .reduceByKey(lambda a, b: a + b) \
        .map(lambda tup: (tup[0][0], {tup[0][1]: tup[1]})) \
        .reduceByKey(lambda a, b: {**a, **b}) \
        .map(lambda tup: (tup[0], [(k, v) for k, v in tup[1].items()])) \
        .map(lambda tup: (tup[0], sorted(tup[1], key=lambda t: (t[1], t[0]), reverse=True))) \
        .map(lambda tup: (tup[0], [t[0] for t in tup[1]])) \
        .collect()
    cat_types = {tup[0]: tup[1] for tup in cat_types}
    con_types = {k: [1.0] for k, v in dtypes.items() if v != 'string'}
    all_types = {**cat_types, **con_types}
    return all_types


# flake8: noqa: F841
def get_columns(formula, sdf, profile=None):
    """
    Gets the expanded columns of the specified Spark dataframe using the specified formula.

    :param formula: Formula (R-like, based on patsy).
    :param sdf: Spark dataframe.
    :param profile: Profile. Default is `None` and profile will be determined empirically.
    :return: Tuple of columns for y, X.
    """
    if profile is None:
        profile = get_profile(sdf)

    data = product(*(v for _, v in profile.items()))
    columns = [k for k, _ in profile.items()]
    df = pd.DataFrame(data, columns=columns)

    if 'np.' in formula:
        np = importlib.import_module('numpy')
    y, X = dmatrices(formula, df, return_type='dataframe')

    return list(y), list(X)


def __smatrices(columns, sdf):
    """
    Constructs new Spark dataframe based on columns.

    :param columns: Columns generated from patsy.
    :param sdf: Spark dataframe.
    :return: Spark dataframe.
    """

    def to_record(record):
        return Row(**{term: InteractionExtractor(record, term).value for term in columns})

    return sdf.rdd \
        .map(lambda r: to_record(r.asDict())) \
        .toDF()


def smatrices(formula, sdf, profile=None):
    """
    Gets tuple of design/model matrices.

    :param formula: Formula.
    :param sdf: Spark dataframe.
    :param profile: Dictionary of data profile.
    :return: y, X Spark dataframes.
    """
    y_cols, X_cols = get_columns(formula, sdf, profile=profile)
    X = __smatrices(X_cols, sdf)
    y = __smatrices(y_cols, sdf)

    return y, X
